@@ -11,6 +11,7 @@ from typing import (
     Callable,
     Dict,
     FrozenSet,
+    Iterable,
     List,
     Optional,
     Sequence,
@@ -44,12 +45,14 @@ from .types import (
     ConstrainedFloat,
     ConstrainedInt,
     ConstrainedList,
+    ConstrainedSet,
     ConstrainedStr,
     conbytes,
     condecimal,
     confloat,
     conint,
     conlist,
+    conset,
     constr,
 )
 from .typing import ForwardRef, Literal, is_callable_type, is_literal_type, literal_values
@@ -62,6 +65,10 @@ if TYPE_CHECKING:
 default_prefix = '#/definitions/'
 
 
+TypeModelOrEnum = Union[Type['BaseModel'], Type[Enum]]
+TypeModelSet = Set[TypeModelOrEnum]
+
+
 def schema(
     models: Sequence[Union[Type['BaseModel'], Type['DataclassType']]],
     *,
@@ -145,9 +152,9 @@ def field_schema(
     field: ModelField,
     *,
     by_alias: bool = True,
-    model_name_map: Dict[Type['BaseModel'], str],
+    model_name_map: Dict[TypeModelOrEnum, str],
     ref_prefix: Optional[str] = None,
-    known_models: Set[Type['BaseModel']] = None,
+    known_models: TypeModelSet = None,
 ) -> Tuple[Dict[str, Any], Dict[str, Any], Set[str]]:
     """
     Process a Pydantic field and return a tuple with a JSON Schema for it as the first item.
@@ -234,10 +241,13 @@ def get_field_schema_validations(field: ModelField) -> Dict[str, Any]:
         f_schema['const'] = field.default
     if field.field_info.extra:
         f_schema.update(field.field_info.extra)
+    modify_schema = getattr(field.outer_type_, '__modify_schema__', None)
+    if modify_schema:
+        modify_schema(f_schema)
     return f_schema
 
 
-def get_model_name_map(unique_models: Set[Type['BaseModel']]) -> Dict[Type['BaseModel'], str]:
+def get_model_name_map(unique_models: TypeModelSet) -> Dict[TypeModelOrEnum, str]:
     """
     Process a set of models and generate unique names for them to be used as keys in the JSON Schema
     definitions. By default the names are the same as the class name. But if two models in different Python
@@ -250,8 +260,7 @@ def get_model_name_map(unique_models: Set[Type['BaseModel']]) -> Dict[Type['Base
     name_model_map = {}
     conflicting_names: Set[str] = set()
     for model in unique_models:
-        model_name = model.__name__
-        model_name = re.sub(r'[^a-zA-Z0-9.\-_]', '_', model_name)
+        model_name = normalize_name(model.__name__)
         if model_name in conflicting_names:
             model_name = get_long_model_name(model)
             name_model_map[model_name] = model
@@ -265,9 +274,7 @@ def get_model_name_map(unique_models: Set[Type['BaseModel']]) -> Dict[Type['Base
     return {v: k for k, v in name_model_map.items()}
 
 
-def get_flat_models_from_model(
-    model: Type['BaseModel'], known_models: Set[Type['BaseModel']] = None
-) -> Set[Type['BaseModel']]:
+def get_flat_models_from_model(model: Type['BaseModel'], known_models: TypeModelSet = None) -> TypeModelSet:
     """
     Take a single ``model`` and generate a set with itself and all the sub-models in the tree. I.e. if you pass
     model ``Foo`` (subclass of Pydantic ``BaseModel``) as ``model``, and it has a field of type ``Bar`` (also
@@ -279,7 +286,7 @@ def get_flat_models_from_model(
     :return: a set with the initial model and all its sub-models
     """
     known_models = known_models or set()
-    flat_models: Set[Type['BaseModel']] = set()
+    flat_models: TypeModelSet = set()
     flat_models.add(model)
     known_models |= flat_models
     fields = cast(Sequence[ModelField], model.__fields__.values())
@@ -287,7 +294,7 @@ def get_flat_models_from_model(
     return flat_models
 
 
-def get_flat_models_from_field(field: ModelField, known_models: Set[Type['BaseModel']]) -> Set[Type['BaseModel']]:
+def get_flat_models_from_field(field: ModelField, known_models: TypeModelSet) -> TypeModelSet:
     """
     Take a single Pydantic ``ModelField`` (from a model) that could have been declared as a sublcass of BaseModel
     (so, it could be a submodel), and generate a set with its model and all the sub-models in the tree.
@@ -301,7 +308,7 @@ def get_flat_models_from_field(field: ModelField, known_models: Set[Type['BaseMo
     """
     from .main import BaseModel  # noqa: F811
 
-    flat_models: Set[Type[BaseModel]] = set()
+    flat_models: TypeModelSet = set()
     # Handle dataclass-based models
     field_type = field.type_
     if lenient_issubclass(getattr(field_type, '__pydantic_model__', None), BaseModel):
@@ -310,12 +317,12 @@ def get_flat_models_from_field(field: ModelField, known_models: Set[Type['BaseMo
         flat_models |= get_flat_models_from_fields(field.sub_fields, known_models=known_models)
     elif lenient_issubclass(field_type, BaseModel) and field_type not in known_models:
         flat_models |= get_flat_models_from_model(field_type, known_models=known_models)
+    elif lenient_issubclass(field_type, Enum):
+        flat_models.add(field_type)
     return flat_models
 
 
-def get_flat_models_from_fields(
-    fields: Sequence[ModelField], known_models: Set[Type['BaseModel']]
-) -> Set[Type['BaseModel']]:
+def get_flat_models_from_fields(fields: Sequence[ModelField], known_models: TypeModelSet) -> TypeModelSet:
     """
     Take a list of Pydantic  ``ModelField``s (from a model) that could have been declared as sublcasses of ``BaseModel``
     (so, any of them could be a submodel), and generate a set with their models and all the sub-models in the tree.
@@ -327,25 +334,25 @@ def get_flat_models_from_fields(
     :param known_models: used to solve circular references
     :return: a set with any model declared in the fields, and all their sub-models
     """
-    flat_models: Set[Type['BaseModel']] = set()
+    flat_models: TypeModelSet = set()
     for field in fields:
         flat_models |= get_flat_models_from_field(field, known_models=known_models)
     return flat_models
 
 
-def get_flat_models_from_models(models: Sequence[Type['BaseModel']]) -> Set[Type['BaseModel']]:
+def get_flat_models_from_models(models: Sequence[Type['BaseModel']]) -> TypeModelSet:
     """
     Take a list of ``models`` and generate a set with them and all their sub-models in their trees. I.e. if you pass
     a list of two models, ``Foo`` and ``Bar``, both subclasses of Pydantic ``BaseModel`` as models, and ``Bar`` has
     a field of type ``Baz`` (also subclass of ``BaseModel``), the return value will be ``set([Foo, Bar, Baz])``.
     """
-    flat_models: Set[Type['BaseModel']] = set()
+    flat_models: TypeModelSet = set()
     for model in models:
         flat_models |= get_flat_models_from_model(model)
     return flat_models
 
 
-def get_long_model_name(model: Type['BaseModel']) -> str:
+def get_long_model_name(model: TypeModelOrEnum) -> str:
     return f'{model.__module__}__{model.__name__}'.replace('.', '__')
 
 
@@ -353,10 +360,10 @@ def field_type_schema(
     field: ModelField,
     *,
     by_alias: bool,
-    model_name_map: Dict[Type['BaseModel'], str],
+    model_name_map: Dict[TypeModelOrEnum, str],
     schema_overrides: bool = False,
     ref_prefix: Optional[str] = None,
-    known_models: Set[Type['BaseModel']],
+    known_models: TypeModelSet,
 ) -> Tuple[Dict[str, Any], Dict[str, Any], Set[str]]:
     """
     Used by ``field_schema()``, you probably should be using that function.
@@ -429,12 +436,12 @@ def field_type_schema(
 
 
 def model_process_schema(
-    model: Type['BaseModel'],
+    model: TypeModelOrEnum,
     *,
     by_alias: bool = True,
-    model_name_map: Dict[Type['BaseModel'], str],
+    model_name_map: Dict[TypeModelOrEnum, str],
     ref_prefix: Optional[str] = None,
-    known_models: Set[Type['BaseModel']] = None,
+    known_models: TypeModelSet = None,
 ) -> Tuple[Dict[str, Any], Dict[str, Any], Set[str]]:
     """
     Used by ``model_schema()``, you probably should be using that function.
@@ -447,6 +454,11 @@ def model_process_schema(
 
     ref_prefix = ref_prefix or default_prefix
     known_models = known_models or set()
+    if lenient_issubclass(model, Enum):
+        model = cast(Type[Enum], model)
+        s = enum_process_schema(model)
+        return s, {}, set()
+    model = cast(Type['BaseModel'], model)
     s = {'title': model.__config__.title or model.__name__}
     doc = getdoc(model)
     if doc:
@@ -471,9 +483,9 @@ def model_type_schema(
     model: Type['BaseModel'],
     *,
     by_alias: bool,
-    model_name_map: Dict[Type['BaseModel'], str],
+    model_name_map: Dict[TypeModelOrEnum, str],
     ref_prefix: Optional[str] = None,
-    known_models: Set[Type['BaseModel']],
+    known_models: TypeModelSet,
 ) -> Tuple[Dict[str, Any], Dict[str, Any], Set[str]]:
     """
     You probably should be using ``model_schema()``, this function is indirectly used by that function.
@@ -516,14 +528,40 @@ def model_type_schema(
     return out_schema, definitions, nested_models
 
 
+def enum_process_schema(enum: Type[Enum]) -> Dict[str, Any]:
+    """
+    Take a single `enum` and generate its schema.
+
+    This is similar to the `model_process_schema` function, but applies to ``Enum`` objects.
+    """
+    from inspect import getdoc
+
+    schema: Dict[str, Any] = {
+        'title': enum.__name__,
+        # Python assigns all enums a default docstring value of 'An enumeration', so
+        # all enums will have a description field even if not explicitly provided.
+        'description': getdoc(enum),
+        # Add enum values and the enum field type to the schema.
+        'enum': [item.value for item in cast(Iterable[Enum], enum)],
+    }
+
+    add_field_type_to_schema(enum, schema)
+
+    modify_schema = getattr(enum, '__modify_schema__', None)
+    if modify_schema:
+        modify_schema(schema)
+
+    return schema
+
+
 def field_singleton_sub_fields_schema(
     sub_fields: Sequence[ModelField],
     *,
     by_alias: bool,
-    model_name_map: Dict[Type['BaseModel'], str],
+    model_name_map: Dict[TypeModelOrEnum, str],
     schema_overrides: bool = False,
     ref_prefix: Optional[str] = None,
-    known_models: Set[Type['BaseModel']],
+    known_models: TypeModelSet,
 ) -> Tuple[Dict[str, Any], Dict[str, Any], Set[str]]:
     """
     This function is indirectly used by ``field_schema()``, you probably should be using that function.
@@ -556,6 +594,11 @@ def field_singleton_sub_fields_schema(
                 known_models=known_models,
             )
             definitions.update(sub_definitions)
+            if schema_overrides and 'allOf' in sub_schema:
+                # if the sub_field is a referenced schema we only need the referenced
+                # object. Otherwise we will end up with several allOf inside anyOf.
+                # See https://github.com/samuelcolvin/pydantic/issues/1209
+                sub_schema = sub_schema['allOf'][0]
             sub_field_schemas.append(sub_schema)
             nested_models.update(sub_nested_models)
         return {'anyOf': sub_field_schemas}, definitions, nested_models
@@ -586,19 +629,33 @@ field_class_to_schema: Tuple[Tuple[Any, Dict[str, Any]], ...] = (
     (list, {'type': 'array', 'items': {}}),
     (tuple, {'type': 'array', 'items': {}}),
     (set, {'type': 'array', 'items': {}, 'uniqueItems': True}),
+    (frozenset, {'type': 'array', 'items': {}, 'uniqueItems': True}),
 )
 
 json_scheme = {'type': 'string', 'format': 'json-string'}
 
 
+def add_field_type_to_schema(field_type: Any, schema: Dict[str, Any]) -> None:
+    """
+    Update the given `schema` with the type-specific metadata for the given `field_type`.
+
+    This function looks through `field_class_to_schema` for a class that matches the given `field_type`,
+    and then modifies the given `schema` with the information from that type.
+    """
+    for type_, t_schema in field_class_to_schema:
+        if issubclass(field_type, type_):
+            schema.update(t_schema)
+            break
+
+
 def field_singleton_schema(  # noqa: C901 (ignore complexity)
     field: ModelField,
     *,
     by_alias: bool,
-    model_name_map: Dict[Type['BaseModel'], str],
+    model_name_map: Dict[TypeModelOrEnum, str],
     schema_overrides: bool = False,
     ref_prefix: Optional[str] = None,
-    known_models: Set[Type['BaseModel']],
+    known_models: TypeModelSet,
 ) -> Tuple[Dict[str, Any], Dict[str, Any], Set[str]]:
     """
     This function is indirectly used by ``field_schema()``, you should probably be using that function.
@@ -641,18 +698,16 @@ def field_singleton_schema(  # noqa: C901 (ignore complexity)
         field_type = literal_value.__class__
         f_schema['const'] = literal_value
 
-    if issubclass(field_type, Enum):
-        f_schema.update({'enum': [item.value for item in field_type]})
-        # Don't return immediately, to allow adding specific types
+    if lenient_issubclass(field_type, Enum):
+        enum_name = normalize_name(field_type.__name__)
+        f_schema = {'$ref': ref_prefix + enum_name}
+        definitions[enum_name] = enum_process_schema(field_type)
+    else:
+        add_field_type_to_schema(field_type, f_schema)
 
-    for type_, t_schema in field_class_to_schema:
-        if issubclass(field_type, type_):
-            f_schema.update(t_schema)
-            break
-
-    modify_schema = getattr(field_type, '__modify_schema__', None)
-    if modify_schema:
-        modify_schema(f_schema)
+        modify_schema = getattr(field_type, '__modify_schema__', None)
+        if modify_schema:
+            modify_schema(f_schema)
 
     if f_schema:
         return f_schema, definitions, nested_models
@@ -688,7 +743,7 @@ def field_singleton_schema(  # noqa: C901 (ignore complexity)
 def multivalue_literal_field_for_schema(values: Tuple[Any, ...], field: ModelField) -> ModelField:
     return ModelField(
         name=field.name,
-        type_=Union[tuple(Literal[value] for value in values)],
+        type_=Union[tuple(Literal[value] for value in values)],  # type: ignore
         class_validators=field.class_validators,
         model_config=field.model_config,
         default=field.default,
@@ -742,7 +797,11 @@ def get_annotation_from_field_info(annotation: Any, field_info: FieldInfo, field
     used_constraints: Set[str] = set()
 
     def go(type_: Any) -> Type[Any]:
-        if is_literal_type(annotation) or isinstance(type_, ForwardRef) or lenient_issubclass(type_, ConstrainedList):
+        if (
+            is_literal_type(annotation)
+            or isinstance(type_, ForwardRef)
+            or lenient_issubclass(type_, (ConstrainedList, ConstrainedSet))
+        ):
             return type_
         origin = getattr(type_, '__origin__', None)
         if origin is not None:
@@ -752,12 +811,16 @@ def get_annotation_from_field_info(annotation: Any, field_info: FieldInfo, field
                 return type_
 
             if origin is Union:
-                return Union[tuple(go(a) for a in args)]
+                return Union[tuple(go(a) for a in args)]  # type: ignore
 
             if issubclass(origin, List) and (field_info.min_items is not None or field_info.max_items is not None):
                 used_constraints.update({'min_items', 'max_items'})
                 return conlist(go(args[0]), min_items=field_info.min_items, max_items=field_info.max_items)
 
+            if issubclass(origin, Set) and (field_info.min_items is not None or field_info.max_items is not None):
+                used_constraints.update({'min_items', 'max_items'})
+                return conset(go(args[0]), min_items=field_info.min_items, max_items=field_info.max_items)
+
             for t in (Tuple, List, Set, FrozenSet, Sequence):
                 if issubclass(origin, t):  # type: ignore
                     return t[tuple(go(a) for a in args)]  # type: ignore
@@ -775,7 +838,7 @@ def get_annotation_from_field_info(annotation: Any, field_info: FieldInfo, field
                 attrs = ('max_length', 'min_length', 'regex')
                 constraint_func = conbytes
             elif issubclass(type_, numeric_types) and not issubclass(
-                type_, (ConstrainedInt, ConstrainedFloat, ConstrainedDecimal, ConstrainedList, bool)
+                type_, (ConstrainedInt, ConstrainedFloat, ConstrainedDecimal, ConstrainedList, ConstrainedSet, bool)
             ):
                 # Is numeric type
                 attrs = ('gt', 'lt', 'ge', 'le', 'multiple_of')
@@ -807,6 +870,13 @@ def get_annotation_from_field_info(annotation: Any, field_info: FieldInfo, field
     return ans
 
 
+def normalize_name(name: str) -> str:
+    """
+    Normalizes the given name. This can be applied to either a model *or* enum.
+    """
+    return re.sub(r'[^a-zA-Z0-9.\-_]', '_', name)
+
+
 class SkipField(Exception):
     """
     Utility exception used to exclude fields from schema.
